# Copyright (c) 2022 DataCore Software Corp. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Classes to invoke DataCore SANsymphony API."""

import copy
import socket
import sys
import uuid

from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import importutils
import suds
from suds import client as suds_client
from suds import plugin
from suds.sax import attribute
from suds.sax import element
from suds import wsdl
from suds import wsse
from suds import xsd
from suds.xsd.doctor import Import
from suds.xsd.doctor import ImportDoctor

from cinder.i18n import _
from cinder import utils as cinder_utils
from cinder.volume.drivers.datacore import exception as datacore_exceptions
from cinder.volume.drivers.datacore import utils as datacore_utils

websocket = importutils.try_import('websocket')

LOG = logging.getLogger(__name__)


class FaultDefinitionsFilter(plugin.DocumentPlugin):
    """Plugin to process the DataCore API WSDL document.

    The document plugin removes fault definitions for callback operations
    from the DataCore API WSDL.
    """

    def parsed(self, context):
        document = context.document
        tns = self._get_tns(document)

        message_qrefs = set()
        for message in self._get_wsdl_messages(document):
            message_qrefs.add((message.get('name'), tns[1]))

        bindings = self._get_wsdl_operation_bindings(document)

        for port_type in self._get_wsdl_port_types(document):
            for operation in self._get_wsdl_operations(port_type):
                self._filter_faults(
                    document, operation, bindings, message_qrefs, tns)

    @staticmethod
    def _get_tns(document):
        target_namespace = document.get('targetNamespace')
        prefix = document.findPrefix(target_namespace) or 'tns'
        return prefix, target_namespace

    @staticmethod
    def _get_wsdl_port_types(document):
        return document.getChildren('portType', wsdl.wsdlns)

    @staticmethod
    def _get_wsdl_operations(port_type):
        return port_type.getChildren('operation', wsdl.wsdlns)

    @staticmethod
    def _get_wsdl_messages(document):
        return document.getChildren('message', wsdl.wsdlns)

    @staticmethod
    def _get_wsdl_operation_bindings(document):
        bindings = []
        for binding in document.getChildren('binding', wsdl.wsdlns):
            operations = {}
            for operation in binding.getChildren('operation', wsdl.wsdlns):
                operations[operation.get('name')] = operation
            bindings.append(operations)
        return bindings

    @staticmethod
    def _filter_faults(document, operation, operation_bindings,
                       message_qrefs, tns):
        filtered_faults = {}
        for fault in operation.getChildren('fault', wsdl.wsdlns):
            fault_message = fault.get('message')
            qref = xsd.qualify(fault_message, document, tns)
            if qref not in message_qrefs:
                filtered_faults[fault.get('name')] = fault
        for fault in filtered_faults.values():
            operation.remove(fault)
        if filtered_faults:
            for binding in operation_bindings:
                filtered_binding_faults = []
                faults = binding[operation.get('name')].getChildren(
                    'fault', wsdl.wsdlns)
                for binding_fault in faults:
                    if binding_fault.get('name') in filtered_faults:
                        filtered_binding_faults.append(binding_fault)
                for binding_fault in filtered_binding_faults:
                    binding[operation.get('name')].remove(binding_fault)


class DataCoreClient(object):
    """DataCore SANsymphony client."""

    API_RETRY_INTERVAL = 10

    DATACORE_EXECUTIVE_PORT = '3794'

    STORAGE_SERVICES = 'IStorageServices'
    STORAGE_SERVICES_BINDING = 'CustomBinding_IStorageServices'

    EXECUTIVE_SERVICE = 'IExecutiveServiceEx'
    EXECUTIVE_SERVICE_BINDING = 'CustomBinding_IExecutiveServiceEx'

    NS_WSA = ('wsa', 'http://www.w3.org/2005/08/addressing')
    WSA_ANONYMOUS = 'http://www.w3.org/2005/08/addressing/anonymous'
    MUST_UNDERSTAND = attribute.Attribute('SOAP-ENV:mustUnderstand', '1')

    # Namespaces that are defined within DataCore API WSDL
    NS_DATACORE_EXECUTIVE = ('http://schemas.datacontract.org/2004/07/'
                             'DataCore.Executive')
    NS_DATACORE_EXECUTIVE_SCSI = ('http://schemas.datacontract.org/2004/07/'
                                  'DataCore.Executive.Scsi')
    NS_DATACORE_EXECUTIVE_ISCSI = ('http://schemas.datacontract.org/2004/07/'
                                   'DataCore.Executive.iSCSI')
    NS_SERIALIZATION_ARRAYS = ('http://schemas.microsoft.com/2003/10/'
                               'Serialization/Arrays')

    # Fully qualified names of objects that are defined within
    # DataCore API WSDL
    O_ACCESS_TOKEN = '{%s}AccessToken' % NS_DATACORE_EXECUTIVE_ISCSI
    O_ARRAY_OF_PERFORMANCE_TYPE = ('{%s}ArrayOfPerformanceType'
                                   % NS_DATACORE_EXECUTIVE)
    O_ARRAY_OF_STRING = '{%s}ArrayOfstring' % NS_SERIALIZATION_ARRAYS
    O_CLIENT_MACHINE_TYPE = '{%s}ClientMachineType' % NS_DATACORE_EXECUTIVE
    O_DATA_SIZE = '{%s}DataSize' % NS_DATACORE_EXECUTIVE
    O_LOGICAL_DISK_ROLE = '{%s}LogicalDiskRole' % NS_DATACORE_EXECUTIVE
    O_LOGICAL_UNIT_TYPE = '{%s}LogicalUnitType' % NS_DATACORE_EXECUTIVE
    O_MIRROR_RECOVERY_PRIORITY = ('{%s}MirrorRecoveryPriority'
                                  % NS_DATACORE_EXECUTIVE)
    O_PATH_POLICY = '{%s}PathPolicy' % NS_DATACORE_EXECUTIVE
    O_PERFORMANCE_TYPE = '{%s}PerformanceType' % NS_DATACORE_EXECUTIVE
    O_POOL_VOLUME_TYPE = '{%s}PoolVolumeType' % NS_DATACORE_EXECUTIVE
    O_SNAPSHOT_TYPE = '{%s}SnapshotType' % NS_DATACORE_EXECUTIVE
    O_SCSI_MODE = '{%s}ScsiMode' % NS_DATACORE_EXECUTIVE_SCSI
    O_SCSI_PORT_DATA = '{%s}ScsiPortData' % NS_DATACORE_EXECUTIVE
    O_SCSI_PORT_NEXUS_DATA = '{%s}ScsiPortNexusData' % NS_DATACORE_EXECUTIVE
    O_SCSI_PORT_TYPE = '{%s}ScsiPortType' % NS_DATACORE_EXECUTIVE_SCSI
    O_VIRTUAL_DISK_DATA = '{%s}VirtualDiskData' % NS_DATACORE_EXECUTIVE
    O_VIRTUAL_DISK_STATUS = '{%s}VirtualDiskStatus' % NS_DATACORE_EXECUTIVE
    O_VIRTUAL_DISK_SUB_TYPE = '{%s}VirtualDiskSubType' % NS_DATACORE_EXECUTIVE
    O_VIRTUAL_DISK_TYPE = '{%s}VirtualDiskType' % NS_DATACORE_EXECUTIVE

    def __init__(self, host, username, password, timeout):
        if websocket is None:
            msg = _("Failed to import websocket-client python module."
                    " Please, ensure the module is installed.")
            raise datacore_exceptions.DataCoreException(msg)

        self.timeout = timeout

        executive_service_net_addr = datacore_utils.build_network_address(
            host, self.DATACORE_EXECUTIVE_PORT)
        executive_service_endpoint = self._build_service_endpoint(
            executive_service_net_addr, self.EXECUTIVE_SERVICE)

        security_options = wsse.Security()
        username_token = wsse.UsernameToken(username, password)
        security_options.tokens.append(username_token)

        imp = Import('http://www.w3.org/2001/XMLSchema',
                     location='http://www.w3.org/2001/XMLSchema.xsd')
        imp.filter.add(
            'http://schemas.microsoft.com/2003/10/Serialization/Arrays')
        imp.filter.add('http://schemas.microsoft.com/2003/10/Serialization/')
        imp.filter.add(
            'http://schemas.datacontract.org/2004/07/DataCore.Executive')

        self._executive_service_client = suds_client.Client(
            executive_service_endpoint['http_endpoint'] + '?singlewsdl',
            nosend=True,
            timeout=self.timeout,
            wsse=security_options,
            plugins=[FaultDefinitionsFilter()],
            doctor=ImportDoctor(imp))

        self._update_storage_services_endpoint(executive_service_endpoint)

        storage_services_endpoint = self._get_storage_services_endpoint()

        self._storage_services_client = suds_client.Client(
            storage_services_endpoint['http_endpoint'] + '?singlewsdl',
            nosend=True,
            timeout=self.timeout,
            wsse=security_options,
            plugins=[FaultDefinitionsFilter()], doctor=ImportDoctor(imp))

        self._update_executive_service_endpoints(storage_services_endpoint)

    @staticmethod
    def _get_list_data(obj, attribute_name):
        return getattr(obj, attribute_name, [])

    @staticmethod
    def _build_service_endpoint(network_address, path):
        return {
            'network_address': network_address,
            'http_endpoint': '%s://%s/%s' % ('http', network_address, path),
            'ws_endpoint': '%s://%s/%s' % ('ws', network_address, path),
        }

    @cinder_utils.synchronized('datacore-api-request_context')
    def _get_soap_context(self, service_client, service_binding, method,
                          message_id, *args, **kwargs):
        soap_action = (service_client.wsdl.services[0].port(service_binding)
                       .methods[method].soap.action)

        soap_headers = self._get_soap_headers(soap_action, message_id)

        service_client.set_options(soapheaders=soap_headers)
        context = service_client.service[service_binding][method](
            *args, **kwargs)

        return context

    def _get_soap_headers(self, soap_action, message_id):
        headers = [
            element.Element('Action', ns=self.NS_WSA).setText(
                soap_action.replace('"', '')).append(self.MUST_UNDERSTAND),

            element.Element('To', ns=self.NS_WSA).setText(
                self.WSA_ANONYMOUS).append(self.MUST_UNDERSTAND),

            element.Element('MessageID', ns=self.NS_WSA).setText(message_id),

            element.Element('ReplyTo', ns=self.NS_WSA).insert(
                element.Element('Address', ns=self.NS_WSA).setText(
                    self.WSA_ANONYMOUS)),
        ]
        return headers

    def _process_request(self, service_client, service_binding,
                         service_endpoint, method, *args, **kwargs):
        max_date = '9999-12-31T23:59:59.9999999'
        r_date = '9999-12-31T23:59:59.9'
        message_id = uuid.uuid4().urn

        context = self._get_soap_context(
            service_client, service_binding,
            method, message_id, *args, **kwargs)

        channel = None
        try:
            channel = websocket.create_connection(
                service_endpoint,
                timeout=self.timeout,
                subprotocols=['soap'],
                header=['soap-content-type: text/xml'])
            channel.send(context.envelope)
            response = channel.recv()
            if not isinstance(response, str):
                response = response.decode('utf-8')
            response = response.replace(max_date, r_date)
            response = response.encode('utf-8')
            return context.process_reply(response)
        except (socket.error, websocket.WebSocketException) as e:
            error = datacore_exceptions.DataCoreConnectionException(reason=e)
            raise error.with_traceback(sys.exc_info()[2])
        except suds.WebFault as e:
            fault = datacore_exceptions.DataCoreFaultException(reason=e)
            raise fault.with_traceback(sys.exc_info()[2])
        finally:
            if channel and channel.connected:
                try:
                    channel.close()
                except (socket.error, websocket.WebSocketException) as e:
                    LOG.debug("Closing a connection to "
                              "DataCore server failed. %s", e)

    def _invoke_storage_services(self, method, *args, **kwargs):

        @cinder_utils.retry(
            datacore_exceptions.DataCoreConnectionException,
            interval=self.API_RETRY_INTERVAL,
            retries=10,
            wait_random=True)
        def retry_call():
            storage_services_endpoint = self._get_storage_services_endpoint()
            try:
                result = self._process_request(
                    self._storage_services_client,
                    self.STORAGE_SERVICES_BINDING,
                    storage_services_endpoint['ws_endpoint'],
                    method, *args, **kwargs)
                return result
            except datacore_exceptions.DataCoreConnectionException:
                with excutils.save_and_reraise_exception():
                    self._update_api_endpoints()

        return retry_call()

    def _update_api_endpoints(self):
        executive_service_endpoints = self._get_executive_service_endpoints()
        for endpoint in executive_service_endpoints:
            try:
                self._update_storage_services_endpoint(endpoint)
                break
            except datacore_exceptions.DataCoreConnectionException as e:
                LOG.warning("Failed to update DataCore Server Group "
                            "endpoints. %s.", e)

        storage_services_endpoint = self._get_storage_services_endpoint()
        try:
            self._update_executive_service_endpoints(
                storage_services_endpoint)
        except datacore_exceptions.DataCoreConnectionException as e:
            LOG.warning("Failed to update DataCore Server Group "
                        "endpoints. %s.", e)

    @cinder_utils.synchronized('datacore-api-storage_services_endpoint')
    def _get_storage_services_endpoint(self):
        if self._storage_services_endpoint:
            return copy.copy(self._storage_services_endpoint)
        return None

    @cinder_utils.synchronized('datacore-api-storage_services_endpoint')
    def _update_storage_services_endpoint(self, executive_service_endpoint):
        controller_address = self._process_request(
            self._executive_service_client,
            self.EXECUTIVE_SERVICE_BINDING,
            executive_service_endpoint['ws_endpoint'],
            'GetControllerAddress')

        if not controller_address:
            msg = _("Could not determine controller node.")
            raise datacore_exceptions.DataCoreConnectionException(reason=msg)

        controller_host = controller_address.rsplit(':', 1)[0].strip('[]')
        controller_net_addr = datacore_utils.build_network_address(
            controller_host,
            self.DATACORE_EXECUTIVE_PORT)

        self._storage_services_endpoint = self._build_service_endpoint(
            controller_net_addr,
            self.STORAGE_SERVICES)

    @cinder_utils.synchronized('datacore-api-executive_service_endpoints')
    def _get_executive_service_endpoints(self):
        if self._executive_service_endpoints:
            return self._executive_service_endpoints[:]
        return []

    @cinder_utils.synchronized('datacore-api-executive_service_endpoints')
    def _update_executive_service_endpoints(self, storage_services_endpoint):
        endpoints = []
        nodes = self._get_list_data(
            self._process_request(self._storage_services_client,
                                  self.STORAGE_SERVICES_BINDING,
                                  storage_services_endpoint['ws_endpoint'],
                                  'GetNodes'),
            'RegionNodeData')

        if not nodes:
            msg = _("Could not determine executive nodes.")
            raise datacore_exceptions.DataCoreConnectionException(reason=msg)

        for node in nodes:
            host = node.HostAddress.rsplit(':', 1)[0].strip('[]')
            endpoint = self._build_service_endpoint(
                datacore_utils.build_network_address(
                    host, self.DATACORE_EXECUTIVE_PORT),
                self.EXECUTIVE_SERVICE)
            endpoints.append(endpoint)

        self._executive_service_endpoints = endpoints

    def get_server_groups(self):
        """Get all the server groups in the configuration.

        :return: A list of server group data.
        """

        return self._get_list_data(
            self._invoke_storage_services('GetServerGroups'),
            'ServerHostGroupData')

    def get_servers(self):
        """Get all the server hosts in the configuration.

        :return: A list of server host data
        """

        return self._get_list_data(
            self._invoke_storage_services('GetServers'),
            'ServerHostData')

    def get_disk_pools(self):
        """Get all the pools in the server group.

        :return: A list of disk pool data
        """

        return self._get_list_data(
            self._invoke_storage_services('GetDiskPools'),
            'DiskPoolData')

    def get_logical_disks(self):
        """Get all the logical disks defined in the system.

        :return: A list of logical disks
        """

        return self._get_list_data(
            self._invoke_storage_services('GetLogicalDisks'),
            'LogicalDiskData')

    def create_pool_logical_disk(self, pool_id, pool_volume_type, size,
                                 min_quota=None, max_quota=None):
        """Create the pool logical disk.

        :param pool_id: Pool id
        :param pool_volume_type: Type, either striped or spanned
        :param size: Size
        :param min_quota: Min quota
        :param max_quota: Max quota
        :return: New logical disk data
        """

        volume_type = getattr(self._storage_services_client.factory
                              .create(self.O_POOL_VOLUME_TYPE),
                              pool_volume_type)

        data_size = (self._storage_services_client.factory
                     .create(self.O_DATA_SIZE))
        data_size.Value = size

        data_size_min_quota = None
        if min_quota:
            data_size_min_quota = (self._storage_services_client.factory
                                   .create(self.O_DATA_SIZE))
            data_size_min_quota.Value = min_quota

        data_size_max_quota = None
        if max_quota:
            data_size_max_quota = (self._storage_services_client.factory
                                   .create(self.O_DATA_SIZE))
            data_size_max_quota.Value = max_quota

        return self._invoke_storage_services('CreatePoolLogicalDisk',
                                             poolId=pool_id,
                                             type=volume_type,
                                             size=data_size,
                                             minQuota=data_size_min_quota,
                                             maxQuota=data_size_max_quota)

    def delete_logical_disk(self, logical_disk_id):
        """Delete the logical disk.

        :param logical_disk_id: Logical disk id
        """

        self._invoke_storage_services('DeleteLogicalDisk',
                                      logicalDiskId=logical_disk_id)

    def get_logical_disk_chunk_allocation_map(self, logical_disk_id):
        """Get the logical disk chunk allocation map.

        The logical disk allocation map details all the physical disk chunks
        that are currently allocated to this logical disk.

        :param logical_disk_id: Logical disk id
        :return: A list of member allocation maps, restricted to chunks
                 allocated on to this logical disk
        """

        return self._get_list_data(
            self._invoke_storage_services('GetLogicalDiskChunkAllocationMap',
                                          logicalDiskId=logical_disk_id),
            'MemberAllocationInfoData')

    def get_next_virtual_disk_alias(self, base_alias):
        """Get the next available (unused) virtual disk alias.

        :param base_alias: Base string of the new alias
        :return: New alias
        """

        return self._invoke_storage_services('GetNextVirtualDiskAlias',
                                             baseAlias=base_alias)

    def get_virtual_disks(self):
        """Get all the virtual disks in the configuration.

        :return: A list of virtual disk's data
        """

        return self._get_list_data(
            self._invoke_storage_services('GetVirtualDisks'),
            'VirtualDiskData')

    def build_virtual_disk_data(self, virtual_disk_alias, virtual_disk_type,
                                size, description, storage_profile_id):
        """Create VirtualDiskData object.

        :param virtual_disk_alias: User-visible alias of the virtual disk,
                                   which must be unique
        :param virtual_disk_type: Virtual disk type
        :param size: Virtual disk size
        :param description: A user-readable description of the virtual disk
        :param storage_profile_id: Virtual disk storage profile
        :return: VirtualDiskData object
        """

        vd_data = (self._storage_services_client.factory
                   .create(self.O_VIRTUAL_DISK_DATA))
        vd_data.Size = (self._storage_services_client.factory
                        .create(self.O_DATA_SIZE))
        vd_data.Size.Value = size
        vd_data.Alias = virtual_disk_alias
        vd_data.Description = description
        vd_data.Type = getattr(self._storage_services_client.factory
                               .create(self.O_VIRTUAL_DISK_TYPE),
                               virtual_disk_type)
        vd_data.SubType = getattr(self._storage_services_client.factory
                                  .create(self.O_VIRTUAL_DISK_SUB_TYPE),
                                  'Standard')
        vd_data.DiskStatus = getattr(self._storage_services_client.factory
                                     .create(self.O_VIRTUAL_DISK_STATUS),
                                     'Online')
        vd_data.RecoveryPriority = getattr(
            self._storage_services_client.factory
                .create(self.O_MIRROR_RECOVERY_PRIORITY),
            'Unset')
        vd_data.StorageProfileId = storage_profile_id

        return vd_data

    def create_virtual_disk_ex2(self, virtual_disk_data, first_logical_disk_id,
                                second_logical_disk_id, add_redundancy):
        """Create a virtual disk specifying both the logical disks.

        :param virtual_disk_data: Virtual disk's properties
        :param first_logical_disk_id: Id of the logical disk to use
        :param second_logical_disk_id: Id of the second logical disk to use
        :param add_redundancy: If True, the mirror has redundant mirror paths
        :return: New virtual disk's data
        """

        return self._invoke_storage_services(
            'CreateVirtualDiskEx2',
            virtualDisk=virtual_disk_data,
            firstLogicalDiskId=first_logical_disk_id,
            secondLogicalDiskId=second_logical_disk_id,
            addRedundancy=add_redundancy)

    def set_virtual_disk_size(self, virtual_disk_id, size):
        """Change the size of a virtual disk.

        :param virtual_disk_id: Id of the virtual disk
        :param size: New size
        :return: Virtual disk's data
        """

        data_size = (self._storage_services_client.factory
                     .create(self.O_DATA_SIZE))
        data_size.Value = size

        return self._invoke_storage_services('SetVirtualDiskSize',
                                             virtualDiskId=virtual_disk_id,
                                             size=data_size)

    def delete_virtual_disk(self, virtual_disk_id, delete_logical_disks):
        """Delete a virtual disk.

        :param virtual_disk_id: Id of the virtual disk
        :param delete_logical_disks: If True, delete the associated
                                     logical disks
        """

        self._invoke_storage_services('DeleteVirtualDisk',
                                      virtualDiskId=virtual_disk_id,
                                      deleteLogicalDisks=delete_logical_disks)

    def serve_virtual_disks_to_host(self, host_id, virtual_disks):
        """Serve multiple virtual disks to a specified host.

        :param host_id: Id of the host machine
        :param virtual_disks: A list of virtual disks to serve
        :return: A list of the virtual disks actually served to the host
        """

        virtual_disk_array = (self._storage_services_client.factory
                              .create(self.O_ARRAY_OF_STRING))
        virtual_disk_array.string = virtual_disks

        return self._get_list_data(
            self._invoke_storage_services('ServeVirtualDisksToHost',
                                          hostId=host_id,
                                          virtualDisks=virtual_disk_array),
            'VirtualLogicalUnitData')

    def unserve_virtual_disks_from_host(self, host_id, virtual_disks):
        """Unserve multiple virtual disks from a specified host.

        :param host_id: Id of the host machine
        :param virtual_disks: A list of virtual disks to unserve
        """

        virtual_disk_array = (self._storage_services_client.factory
                              .create(self.O_ARRAY_OF_STRING))
        virtual_disk_array.string = virtual_disks

        self._invoke_storage_services('UnserveVirtualDisksFromHost',
                                      hostId=host_id,
                                      virtualDisks=virtual_disk_array)

    def unserve_virtual_disks_from_port(self, port_id, virtual_disks):
        """Unserve multiple virtual disks from a specified initiator port.

        :param port_id: Id of the initiator port
        :param virtual_disks: A list of virtual disks to unserve
        """

        virtual_disk_array = (self._storage_services_client.factory
                              .create(self.O_ARRAY_OF_STRING))
        virtual_disk_array.string = virtual_disks

        self._invoke_storage_services('UnserveVirtualDisksFromPort',
                                      portId=port_id,
                                      virtualDisks=virtual_disk_array)

    def bind_logical_disk(self, virtual_disk_id, logical_disk_id, role,
                          create_mirror_mappings, create_client_mappings,
                          add_redundancy):
        """Bind (add) a logical disk to a virtual disk.

        :param virtual_disk_id: Id of the virtual disk to bind to
        :param logical_disk_id: Id of the logical disk being bound
        :param role: logical disk's role
        :param create_mirror_mappings: If True, automatically create the
                                       mirror mappings to this disk, assuming
                                       there is already another logical disk
                                       bound
        :param create_client_mappings: If True, automatically create mappings
                                       from mapped hosts to the new disk
        :param add_redundancy: If True, the mirror has redundant mirror paths
        :return: Updated virtual disk data
        """

        logical_disk_role = getattr(self._storage_services_client.factory
                                    .create(self.O_LOGICAL_DISK_ROLE),
                                    role)

        return self._invoke_storage_services(
            'BindLogicalDisk',
            virtualDiskId=virtual_disk_id,
            logicalDiskId=logical_disk_id,
            role=logical_disk_role,
            createMirrorMappings=create_mirror_mappings,
            createClientMappings=create_client_mappings,
            addRedundancy=add_redundancy)

    def get_snapshots(self):
        """Get all the snapshots on all the servers in the region.

        :return: A list of snapshot data.
        """

        return self._get_list_data(
            self._invoke_storage_services('GetSnapshots'),
            'SnapshotData')

    def create_snapshot(self, virtual_disk_id, name, description,
                        destination_pool_id, snapshot_type,
                        duplicate_disk_id, storage_profile_id):
        """Create a snapshot relationship.

        :param virtual_disk_id: Virtual disk id
        :param name: Name of snapshot
        :param description: Description
        :param destination_pool_id: Destination pool id
        :param snapshot_type: Type of snapshot
        :param duplicate_disk_id: If set to True then the destination virtual
                                  disk's SCSI id will be a duplicate of the
                                  source's
        :param storage_profile_id: Specifies the destination virtual disk's
                                   storage profile
        :return: New snapshot data
        """

        st_type = getattr(self._storage_services_client.factory
                          .create(self.O_SNAPSHOT_TYPE),
                          snapshot_type)

        return self._invoke_storage_services(
            'CreateSnapshot',
            virtualDiskId=virtual_disk_id,
            name=name,
            description=description,
            destinationPoolId=destination_pool_id,
            type=st_type,
            duplicateDiskId=duplicate_disk_id,
            storageProfileId=storage_profile_id)

    def delete_snapshot(self, snapshot_id):
        """Delete the snapshot.

        :param snapshot_id: Snapshot id
        """

        self._invoke_storage_services('DeleteSnapshot', snapshotId=snapshot_id)

    def get_storage_profiles(self):
        """Get all the all the defined storage profiles.

        :return: A list of storage profiles
        """

        return self._get_list_data(
            self._invoke_storage_services('GetStorageProfiles'),
            'StorageProfileData')

    def designate_map_store(self, pool_id):
        """Designate which pool the snapshot mapstore will be allocated from.

        :param pool_id: Pool id
        :return: Updated server host data, which includes the mapstore pool id
        """

        return self._invoke_storage_services('DesignateMapStore',
                                             poolId=pool_id)

    def get_performance_by_type(self, performance_types):
        """Get performance data for specific types of performance counters.

        :param performance_types: A list of performance counter types
        :return: A list of performance data points
        """

        prfm_type_array = (self._storage_services_client.factory
                           .create(self.O_ARRAY_OF_PERFORMANCE_TYPE))
        prfm_type_array.PerformanceType = list(
            getattr(self._storage_services_client.factory
                    .create(self.O_PERFORMANCE_TYPE),
                    performance_type)
            for performance_type in performance_types)

        return self._get_list_data(
            self._invoke_storage_services('GetPerformanceByType',
                                          types=prfm_type_array),
            'CollectionPointData')

    def get_ports(self):
        """Get all ports in the configuration.

        :return: A list of SCSI ports
        """

        return self._get_list_data(
            self._invoke_storage_services('GetPorts'),
            'ScsiPortData')

    def build_scsi_port_data(self, host_id, port_name, port_mode, port_type):
        """Create ScsiPortData object that represents SCSI port, of any type.

        :param host_id: Id of the port's host computer
        :param port_name: Unique name of the port.
        :param port_mode: Mode of port: initiator or target
        :param port_type: Type of port, Fc, iSCSI or loopback
        :return: ScsiPortData object
        """

        scsi_port_data = (self._storage_services_client.factory
                          .create(self.O_SCSI_PORT_DATA))
        scsi_port_data.HostId = host_id
        scsi_port_data.PortName = port_name
        scsi_port_data.PortMode = getattr(self._storage_services_client.factory
                                          .create(self.O_SCSI_MODE),
                                          port_mode)
        scsi_port_data.PortType = getattr(self._storage_services_client.factory
                                          .create(self.O_SCSI_PORT_TYPE),
                                          port_type)

        return scsi_port_data

    def register_port(self, scsi_port_data):
        """Register a port in the configuration.

        :param scsi_port_data: Port data
        :return: Updated port data
        """

        return self._invoke_storage_services('RegisterPort',
                                             port=scsi_port_data)

    def assign_port(self, client_id, port_id):
        """Assign a port to a client.

        :param client_id: Client id
        :param port_id: Port id
        :return: Updated port data,
                 which will now have its host id set to the client id
        """

        return self._invoke_storage_services('AssignPort',
                                             clientId=client_id,
                                             portId=port_id)

    def set_server_port_properties(self, port_id, properties):
        """Set a server port's properties.

        :param port_id: Port id
        :param properties: New properties
        :return: Updated port data
        """

        return self._invoke_storage_services('SetServerPortProperties',
                                             portId=port_id,
                                             properties=properties)

    def build_access_token(self, initiator_node_name, initiator_username,
                           initiator_password, mutual_authentication,
                           target_username, target_password):
        """Create an AccessToken object.

        :param initiator_node_name: Initiator node name
        :param initiator_username: Initiator user name
        :param initiator_password: Initiator password
        :param mutual_authentication: If True the target and the initiator
                                      authenticate each other.
                                      A separate secret is set for each target
                                      and for each initiator in the storage
                                      area network (SAN).
        :param target_username: Target user name
        :param target_password: Target password
        :return: AccessToken object
        """

        access_token = (self._storage_services_client.factory
                        .create(self.O_ACCESS_TOKEN))
        access_token.InitiatorNodeName = initiator_node_name
        access_token.InitiatorUsername = initiator_username
        access_token.InitiatorPassword = initiator_password
        access_token.MutualAuthentication = mutual_authentication
        access_token.TargetUsername = target_username
        access_token.TargetPassword = target_password

        return access_token

    def set_access_token(self, iscsi_port_id, access_token):
        """Set the access token.

        The access token allows access to a specific network node
        from a specific iSCSI port.

        :param iscsi_port_id: Id of the initiator iSCSI port
        :param access_token: Access token to be validated
        :return: Port data
        """

        return self._invoke_storage_services('SetAccessToken',
                                             iScsiPortId=iscsi_port_id,
                                             inputToken=access_token)

    def get_clients(self):
        """Get all the clients in the configuration.

        :return: A list of client data
        """

        return self._get_list_data(
            self._invoke_storage_services('GetClients'),
            'ClientHostData')

    def register_client(self, host_name, description, machine_type,
                        mode, preferred_server_ids):
        """Register the client, creating a client object in the configuration.

        :param host_name: Name of the client
        :param description: Description
        :param machine_type: Type of client
        :param mode: Path policy mode of the client
        :param preferred_server_ids: Preferred server ids
        :return: New client data
        """

        client_machine_type = getattr(self._storage_services_client.factory
                                      .create(self.O_CLIENT_MACHINE_TYPE),
                                      machine_type)
        client_mode = getattr(self._storage_services_client.factory
                              .create(self.O_PATH_POLICY),
                              mode)

        return self._invoke_storage_services(
            'RegisterClient',
            hostName=host_name,
            description=description,
            type=client_machine_type,
            mode=client_mode,
            preferredServerIds=preferred_server_ids)

    def set_client_capabilities(self, client_id, mpio, alua):
        """Set the client capabilities for MPIO and ALUA.

        :param client_id: Client id
        :param mpio: If set to True then MPIO-capable
        :param alua: If set to True then ALUA-capable
        :return: Updated client data
        """

        return self._invoke_storage_services('SetClientCapabilities',
                                             clientId=client_id,
                                             mpio=mpio,
                                             alua=alua)

    def get_target_domains(self):
        """Get all the target domains in the configuration.

        :return: A list of target domains
        """

        return self._get_list_data(
            self._invoke_storage_services('GetTargetDomains'),
            'VirtualTargetDomainData')

    def create_target_domain(self, initiator_host_id, target_host_id):
        """Create a target domain given a pair of hosts, target and initiator.

        :param initiator_host_id: Id of the initiator host machine
        :param target_host_id: Id of the target host server
        :return: New target domain
        """

        return self._invoke_storage_services('CreateTargetDomain',
                                             initiatorHostId=initiator_host_id,
                                             targetHostId=target_host_id)

    def delete_target_domain(self, target_domain_id):
        """Delete a target domain.

        :param target_domain_id: Target domain id
        """

        self._invoke_storage_services('DeleteTargetDomain',
                                      targetDomainId=target_domain_id)

    def get_target_devices(self):
        """Get all the target devices in the configuration.

        :return: A list of target devices
        """

        return self._get_list_data(
            self._invoke_storage_services('GetTargetDevices'),
            'VirtualTargetDeviceData')

    def build_scsi_port_nexus_data(self, initiator_port_id, target_port_id):
        """Create a ScsiPortNexusData object.

        Nexus is a pair of ports that can communicate, one being the initiator,
        the other the target

        :param initiator_port_id: Id of the initiator port
        :param target_port_id: Id of the target port
        :return: ScsiPortNexusData object
        """

        scsi_port_nexus_data = (self._storage_services_client.factory
                                .create(self.O_SCSI_PORT_NEXUS_DATA))
        scsi_port_nexus_data.InitiatorPortId = initiator_port_id
        scsi_port_nexus_data.TargetPortId = target_port_id

        return scsi_port_nexus_data

    def create_target_device(self, target_domain_id, nexus):
        """Create a target device, given a target domain and a nexus.

        :param target_domain_id: Target domain id
        :param nexus: Nexus, or pair of ports
        :return: New target device
        """

        return self._invoke_storage_services('CreateTargetDevice',
                                             targetDomainId=target_domain_id,
                                             nexus=nexus)

    def delete_target_device(self, target_device_id):
        """Delete a target device.

        :param target_device_id: Target device id
        """

        self._invoke_storage_services('DeleteTargetDevice',
                                      targetDeviceId=target_device_id)

    def get_next_free_lun(self, target_device_id):
        """Find the next unused LUN number for a specified target device.

        :param target_device_id: Target device id
        :return: LUN number
        """

        return self._invoke_storage_services('GetNextFreeLun',
                                             targetDeviceId=target_device_id)

    def get_logical_units(self):
        """Get all the mappings configured in the system.

        :return: A list of mappings
        """

        return self._get_list_data(
            self._invoke_storage_services('GetLogicalUnits'),
            'VirtualLogicalUnitData')

    def map_logical_disk(self, logical_disk_id, nexus, lun,
                         initiator_host_id, mapping_type):
        """Map a logical disk to a host.

        :param logical_disk_id: Id of the logical disk
        :param nexus: Nexus, or pair of ports
        :param lun: Logical Unit Number
        :param initiator_host_id: Id of the initiator host machine
        :param mapping_type: Type of mapping
        :return: New mapping
        """

        logical_unit_type = getattr(self._storage_services_client.factory
                                    .create(self.O_LOGICAL_UNIT_TYPE),
                                    mapping_type)

        return self._invoke_storage_services('MapLogicalDisk',
                                             logicalDiskId=logical_disk_id,
                                             nexus=nexus,
                                             lun=lun,
                                             initiatorHostId=initiator_host_id,
                                             mappingType=logical_unit_type)

    def unmap_logical_disk(self, logical_disk_id, nexus):
        """Unmap a logical disk mapped with a specified nexus.

        :param logical_disk_id: Id of the logical disk
        :param nexus: Nexus, or pair of ports
        """

        self._invoke_storage_services('UnmapLogicalDisk',
                                      logicalDiskId=logical_disk_id,
                                      nexusData=nexus)



# Copyright (c) 2017-2020 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Configuration options for Dell EMC PowerFlex (formerly
named Dell EMC VxFlex OS).
"""

from oslo_config import cfg

# deprecated options
VXFLEXOS_REST_SERVER_PORT = "vxflexos_rest_server_port"
VXFLEXOS_ROUND_VOLUME_CAPACITY = "vxflexos_round_volume_capacity"
VXFLEXOS_UNMAP_VOLUME_BEFORE_DELETION = "vxflexos_unmap_volume_before_deletion"
VXFLEXOS_STORAGE_POOLS = "vxflexos_storage_pools"
VXFLEXOS_SERVER_API_VERSION = "vxflexos_server_api_version"
VXFLEXOS_MAX_OVER_SUBSCRIPTION_RATIO = "vxflexos_max_over_subscription_ratio"
VXFLEXOS_ALLOW_NON_PADDED_VOLUMES = "vxflexos_allow_non_padded_volumes"
VXFLEXOS_ALLOW_MIGRATION_DURING_REBUILD = (
    "vxflexos_allow_migration_during_rebuild")

# actual options
POWERFLEX_REST_SERVER_PORT = "powerflex_rest_server_port"
POWERFLEX_ROUND_VOLUME_CAPACITY = "powerflex_round_volume_capacity"
POWERFLEX_UNMAP_VOLUME_BEFORE_DELETION = (
    "powerflex_unmap_volume_before_deletion")
POWERFLEX_STORAGE_POOLS = "powerflex_storage_pools"
POWERFLEX_SERVER_API_VERSION = "powerflex_server_api_version"
POWERFLEX_MAX_OVER_SUBSCRIPTION_RATIO = "powerflex_max_over_subscription_ratio"
POWERFLEX_ALLOW_NON_PADDED_VOLUMES = "powerflex_allow_non_padded_volumes"
POWERFLEX_ALLOW_MIGRATION_DURING_REBUILD = (
    "powerflex_allow_migration_during_rebuild")

deprecated_opts = [
    cfg.PortOpt(VXFLEXOS_REST_SERVER_PORT,
                default=443,
                help='renamed to %s.' %
                     POWERFLEX_REST_SERVER_PORT,
                deprecated_for_removal=True,
                deprecated_reason='Replaced by %s.' %
                                  POWERFLEX_REST_SERVER_PORT),
    cfg.BoolOpt(VXFLEXOS_ROUND_VOLUME_CAPACITY,
                default=True,
                help='renamed to %s.' %
                     POWERFLEX_ROUND_VOLUME_CAPACITY,
                deprecated_for_removal=True,
                deprecated_reason='Replaced by %s.' %
                                  POWERFLEX_ROUND_VOLUME_CAPACITY),
    cfg.BoolOpt(VXFLEXOS_UNMAP_VOLUME_BEFORE_DELETION,
                default=False,
                help='renamed to %s.' %
                     POWERFLEX_ROUND_VOLUME_CAPACITY,
                deprecated_for_removal=True,
                deprecated_reason='Replaced by %s.' %
                                  POWERFLEX_ROUND_VOLUME_CAPACITY),
    cfg.StrOpt(VXFLEXOS_STORAGE_POOLS,
               help='renamed to %s.' %
                    POWERFLEX_STORAGE_POOLS,
               deprecated_for_removal=True,
               deprecated_reason='Replaced by %s.' %
                                 POWERFLEX_STORAGE_POOLS),
    cfg.StrOpt(VXFLEXOS_SERVER_API_VERSION,
               help='renamed to %s.' %
                    POWERFLEX_SERVER_API_VERSION,
               deprecated_for_removal=True,
               deprecated_reason='Replaced by %s.' %
                                 POWERFLEX_SERVER_API_VERSION),
    cfg.FloatOpt(VXFLEXOS_MAX_OVER_SUBSCRIPTION_RATIO,
                 # This option exists to provide a default value for the
                 # PowerFlex driver which is different than the global default.
                 default=10.0,
                 help='renamed to %s.' %
                      POWERFLEX_MAX_OVER_SUBSCRIPTION_RATIO,
                 deprecated_for_removal=True,
                 deprecated_reason='Replaced by %s.' %
                                   POWERFLEX_MAX_OVER_SUBSCRIPTION_RATIO),
    cfg.BoolOpt(VXFLEXOS_ALLOW_NON_PADDED_VOLUMES,
                default=False,
                help='renamed to %s.' %
                     POWERFLEX_ALLOW_NON_PADDED_VOLUMES,
                deprecated_for_removal=True,
                deprecated_reason='Replaced by %s.' %
                                  POWERFLEX_ALLOW_NON_PADDED_VOLUMES),
    cfg.BoolOpt(VXFLEXOS_ALLOW_MIGRATION_DURING_REBUILD,
                default=False,
                help='renamed to %s.' %
                     POWERFLEX_ALLOW_MIGRATION_DURING_REBUILD,
                deprecated_for_removal=True,
                deprecated_reason='Replaced by %s.' %
                                  POWERFLEX_ALLOW_MIGRATION_DURING_REBUILD),
]

actual_opts = [
    cfg.PortOpt(POWERFLEX_REST_SERVER_PORT,
                default=443,
                help='Gateway REST server port.',
                deprecated_name=VXFLEXOS_REST_SERVER_PORT),
    cfg.BoolOpt(POWERFLEX_ROUND_VOLUME_CAPACITY,
                default=True,
                help='Round volume sizes up to 8GB boundaries. '
                     'PowerFlex/VxFlex OS requires volumes to be sized '
                     'in multiples of 8GB. If set to False, volume '
                     'creation will fail for volumes not sized properly',
                deprecated_name=VXFLEXOS_ROUND_VOLUME_CAPACITY
                ),
    cfg.BoolOpt(POWERFLEX_UNMAP_VOLUME_BEFORE_DELETION,
                default=False,
                help='Unmap volumes before deletion.',
                deprecated_name=VXFLEXOS_UNMAP_VOLUME_BEFORE_DELETION),
    cfg.StrOpt(POWERFLEX_STORAGE_POOLS,
               help='Storage Pools. Comma separated list of storage '
                    'pools used to provide volumes. Each pool should '
                    'be specified as a '
                    'protection_domain_name:storage_pool_name value',
               deprecated_name=VXFLEXOS_STORAGE_POOLS),
    cfg.StrOpt(POWERFLEX_SERVER_API_VERSION,
               help='PowerFlex/ScaleIO API version. This value should be '
                    'left as the default value unless otherwise instructed '
                    'by technical support.',
               deprecated_name=VXFLEXOS_SERVER_API_VERSION),
    cfg.FloatOpt(POWERFLEX_MAX_OVER_SUBSCRIPTION_RATIO,
                 # This option exists to provide a default value for the
                 # PowerFlex driver which is different than the global default.
                 default=10.0,
                 help='max_over_subscription_ratio setting for the driver. '
                      'Maximum value allowed is 10.0.',
                 deprecated_name=VXFLEXOS_MAX_OVER_SUBSCRIPTION_RATIO),
    cfg.BoolOpt(POWERFLEX_ALLOW_NON_PADDED_VOLUMES,
                default=False,
                help='Allow volumes to be created in Storage Pools '
                     'when zero padding is disabled. This option should '
                     'not be enabled if multiple tenants will utilize '
                     'volumes from a shared Storage Pool.',
                deprecated_name=VXFLEXOS_ALLOW_NON_PADDED_VOLUMES),
    cfg.BoolOpt(POWERFLEX_ALLOW_MIGRATION_DURING_REBUILD,
                default=False,
                help='Allow volume migration during rebuild.',
                deprecated_name=VXFLEXOS_ALLOW_MIGRATION_DURING_REBUILD),
]



# Copyright 2020 toyou Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
acs5000 iSCSI driver
"""

from oslo_log import log as logging

from cinder.common import constants
from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder import utils
from cinder.volume.drivers.toyou.acs5000 import acs5000_common

LOG = logging.getLogger(__name__)


@interface.volumedriver
class Acs5000ISCSIDriver(acs5000_common.Acs5000CommonDriver):
    """TOYOU ACS5000 storage iSCSI volume driver.

    .. code-block:: none

      Version history:
          1.0.0 - Initial driver

    """

    VENDOR = 'TOYOU'
    VERSION = '1.0.0'
    PROTOCOL = constants.ISCSI

    # ThirdPartySystems wiki page
    CI_WIKI_NAME = 'TOYOU_ACS5000_CI'

    def __init__(self, *args, **kwargs):
        super(Acs5000ISCSIDriver, self).__init__(*args, **kwargs)
        self.protocol = self.PROTOCOL

    @staticmethod
    def get_driver_options():
        return acs5000_common.Acs5000CommonDriver.get_driver_options()

    def validate_connector(self, connector):
        """Check connector for at least one enabled iSCSI protocol."""
        if 'initiator' not in connector:
            LOG.error('The connector does not '
                      'contain the required information.')
            raise exception.InvalidConnectorException(
                missing='initiator')

    @utils.synchronized('acs5000A-host', external=True)
    def initialize_connection(self, volume, connector):
        LOG.debug('enter: initialize_connection: volume '
                  '%(vol)s with connector %(conn)s',
                  {'vol': volume.id, 'conn': connector})
        volume_name = self._convert_name(volume.name)
        ret = self._cmd.create_lun_map(volume_name,
                                       self.protocol,
                                       connector['initiator'])
        if ret['key'] == 0:
            lun_required = ['iscsi_name', 'portal', 'lun']
            lun_info = ret['arr']
            for param in lun_required:
                if param not in lun_info:
                    msg = (_('initialize_connection: Param %(param)s '
                             'was not returned correctly when volume '
                             '%(vol)s mapping.') % {'param': param,
                                                    'vol': volume.id})
                    raise exception.VolumeBackendAPIException(data=msg)
            data = {'target_discovered': False,
                    'target_iqns': lun_info['iscsi_name'],
                    'target_portals': lun_info['portal'],
                    'target_luns': lun_info['lun'],
                    'volume_id': volume.id}
            LOG.debug('leave: initialize_connection: volume '
                      '%(vol)s with connector %(conn)s',
                      {'vol': volume.id, 'conn': connector})
            return {'driver_volume_type': 'iscsi', 'data': data}
        if ret['key'] == 303:
            raise exception.VolumeNotFound(volume_id=volume_name)
        elif ret['key'] == 402:
            raise exception.ISCSITargetAttachFailed(volume_id=volume_name)
        else:
            msg = (_('failed to map the volume %(vol)s to '
                     'connector %(conn)s.') %
                   {'vol': volume['id'], 'conn': connector})
            raise exception.VolumeBackendAPIException(data=msg)

    @utils.synchronized('acs5000A-host', external=True)
    def terminate_connection(self, volume, connector, **kwargs):
        LOG.debug('enter: terminate_connection: volume '
                  '%(vol)s with connector %(conn)s',
                  {'vol': volume.id, 'conn': connector})
        name = self._convert_name(volume.name)
        # -1 means all lun maps
        initiator = '-1'
        if connector and connector['initiator']:
            initiator = connector['initiator']
        if self._check_multi_attached(volume, connector) < 2:
            self._cmd.delete_lun_map(name,
                                     self.protocol,
                                     initiator)
        else:
            LOG.warning('volume %s has been mapped to multi VMs, and these '
                        'VMs belong to the same host. The mapping '
                        'cancellation request is aborted.', volume.id)
        LOG.debug('leave: terminate_connection: volume '
                  '%(vol)s with connector %(conn)s',
                  {'vol': volume.id, 'conn': connector})
        return {'driver_volume_type': 'iscsi', 'data': {}}



# Copyright (c) 2019 FUJITSU LIMITED
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

RAIDGROUP = 2
TPPOOL = 5
SNAPOPC = 4
OPC = 5
RETURN_TO_RESOURCEPOOL = 19
DETACH = 8
BROKEN = 5

JOB_RETRIES = 60
JOB_INTERVAL_SEC = 10
TIMES_MIN = 3
EC_REC = 3
RETRY_INTERVAL = 5
# Error code keyword.
VOLUME_IS_BUSY = 32786
DEVICE_IS_BUSY = 32787
VOLUMENAME_IN_USE = 32788
COPYSESSION_NOT_EXIST = 32793
LUNAME_IN_USE = 4102
LUNAME_NOT_EXIST = 4097  # Only for InvokeMethod(HidePaths).
VOL_PREFIX = "FJosv_"
REPL = "FUJITSU_ReplicationService"
STOR_CONF = "FUJITSU_StorageConfigurationService"
CTRL_CONF = "FUJITSU_ControllerConfigurationService"
UNDEF_MSG = 'Undefined Error!!'

POOL_TYPE_dic = {
    RAIDGROUP: 'RAID_GROUP',
    TPPOOL: 'Thinporvisioning_POOL',
}
POOL_TYPE_list = [
    'RAID',
    'TPP'
]
OPERATION_dic = {
    SNAPOPC: RETURN_TO_RESOURCEPOOL,
    OPC: DETACH,
    EC_REC: DETACH,
}

RETCODE_dic = {
    '0': 'Success',
    '1': 'Method Not Supported',
    '4': 'Failed',
    '5': 'Invalid Parameter',
    '4096': 'Method Parameters Checked - Job Started',
    '4097': 'Size Not Supported',
    '4101': 'Target/initiator combination already exposed',
    '4102': 'Requested logical unit number in use',
    '32769': 'Maximum number of Logical Volume in a RAID group '
             'has been reached',
    '32770': 'Maximum number of Logical Volume in the storage device '
             'has been reached',
    '32771': 'Maximum number of registered Host WWN '
             'has been reached',
    '32772': 'Maximum number of affinity group has been reached',
    '32773': 'Maximum number of host affinity has been reached',
    '32781': 'Not available under current system configuration',
    '32782': 'Controller firmware update in process',
    '32785': 'The RAID group is in busy state',
    '32786': 'The Logical Volume is in busy state',
    '32787': 'The device is in busy state',
    '32788': 'Element Name is in use',
    '32791': 'Maximum number of copy session has been reached',
    '32792': 'No Copy License',
    '32793': 'Session does not exist',
    '32794': 'Phase is not correct',
    '32796': 'Quick Format Error',
    '32801': 'The CA port is in invalid setting',
    '32802': 'The Logical Volume is Mainframe volume',
    '32803': 'The RAID group is not operative',
    '32804': 'The Logical Volume is not operative',
    '32805': 'The Logical Element is Thin provisioning Pool Volume',
    '32806': 'The Logical Volume is pool for copy volume',
    '32807': 'The Logical Volume is unknown volume',
    '32808': 'No Thin Provisioning License',
    '32809': 'The Logical Element is ODX volume',
    '32810': 'The specified volume is under use as NAS volume',
    '32811': 'This operation cannot be performed to the NAS resources',
    '32812': 'This operation cannot be performed to the '
             'Transparent Failover resources',
    '32813': 'This operation cannot be performed to the '
             'VVOL resources',
    '32816': 'Generic fatal error',
    '32817': 'Inconsistent State with 1Step Restore '
             'operation',
    '32818': 'REC Path failure during copy',
    '32819': 'RAID failure during EC/REC copy',
    '32820': 'Previous command in process',
    '32821': 'Cascade local copy session exist',
    '32822': 'Cascade EC/REC session is not suspended',
    '35302': 'Invalid LogicalElement',
    '35304': 'LogicalElement state error',
    '35316': 'Multi-hop error',
    '35318': 'Maximum number of multi-hop has been reached',
    '35324': 'RAID is broken',
    '35331': 'Maximum number of session has been reached(per device)',
    '35333': 'Maximum number of session has been reached(per SourceElement)',
    '35334': 'Maximum number of session has been reached(per TargetElement)',
    '35335': 'Maximum number of Snapshot generation has been reached '
             '(per SourceElement)',
    '35346': 'Copy table size is not setup',
    '35347': 'Copy table size is not enough',
}



# Copyright (c) 2011 Zadara Storage Inc.
# Copyright (c) 2011 OpenStack Foundation
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright (c) 2010 Citrix Systems, Inc.
# Copyright 2011 Ken Pepple
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Built-in volume type properties."""

from __future__ import annotations

from typing import Any, Iterable, Optional, Union   # noqa: H301

from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_log import log as logging
from oslo_utils import uuidutils

from cinder import context
from cinder import db
from cinder import exception
from cinder.i18n import _
from cinder import quota
from cinder import rpc
from cinder import utils

CONF = cfg.CONF
LOG = logging.getLogger(__name__)
QUOTAS = quota.QUOTAS
ENCRYPTION_IGNORED_FIELDS = ('volume_type_id', 'created_at', 'updated_at',
                             'deleted_at', 'encryption_id')
QOS_IGNORED_FIELDS = ('id', 'name', 'created_at', 'updated_at', 'deleted_at')
DEFAULT_VOLUME_TYPE = "__DEFAULT__"

MIN_SIZE_KEY = "provisioning:min_vol_size"
MAX_SIZE_KEY = "provisioning:max_vol_size"


def create(context: context.RequestContext,
           name: str,
           extra_specs: Optional[dict[str, Any]] = None,
           is_public: bool = True,
           projects: Optional[list[str]] = None,
           description: Optional[str] = None):
    """Creates volume types."""
    extra_specs = extra_specs or {}
    projects = projects or []
    elevated = context if context.is_admin else context.elevated()
    try:
        type_ref = db.volume_type_create(elevated,
                                         dict(name=name,
                                              extra_specs=extra_specs,
                                              is_public=is_public,
                                              description=description),
                                         projects=projects)
    except db_exc.DBError:
        LOG.exception('DB error:')
        raise exception.VolumeTypeCreateFailed(name=name,
                                               extra_specs=extra_specs)
    return type_ref


def update(context: context.RequestContext,
           id: Optional[str],
           name: Optional[str],
           description: Optional[str],
           is_public: Optional[bool] = None) -> None:
    """Update volume type by id."""
    if id is None:
        msg = _("id cannot be None")
        raise exception.InvalidVolumeType(reason=msg)
    elevated = context if context.is_admin else context.elevated()
    old_volume_type = get_volume_type(elevated, id)
    try:
        db.volume_type_update(elevated, id,
                              dict(name=name, description=description,
                                   is_public=is_public))
        # Rename resource in quota if volume type name is changed.
        if name:
            old_type_name = old_volume_type.get('name')
            if old_type_name != name:
                old_description = old_volume_type.get('description')
                old_public = old_volume_type.get('is_public')
                try:
                    QUOTAS.update_quota_resource(elevated,
                                                 old_type_name,
                                                 name)
                # Rollback the updated information to the original
                except db_exc.DBError:
                    db.volume_type_update(elevated, id,
                                          dict(name=old_type_name,
                                               description=old_description,
                                               is_public=old_public))
                    raise
    except db_exc.DBError:
        LOG.exception('DB error:')
        raise exception.VolumeTypeUpdateFailed(id=id)


def destroy(context: context.RequestContext, id: str) -> dict[str, Any]:
    """Marks volume types as deleted.

    There must exist at least one volume type (i.e. the default type) in
    the deployment.
    This method achieves that by ensuring:
    1) the default_volume_type is set and is a valid one
    2) the type requested to delete isn't the default type

    :raises VolumeTypeDefaultDeletionError: when the type requested to
                                            delete is the default type
    """
    if id is None:
        msg = _("id cannot be None")
        raise exception.InvalidVolumeType(reason=msg)

    projects_with_default_type = db.get_all_projects_with_default_type(
        context.elevated(), id)
    if len(projects_with_default_type) > 0:
        # don't allow delete if the type requested is a project default
        project_list = [p.project_id for p in projects_with_default_type]
        LOG.exception('Default type with %(volume_type_id)s is associated '
                      'with projects %(projects)s',
                      {'volume_type_id': id,
                       'projects': project_list})
        raise exception.VolumeTypeDefaultDeletionError(volume_type_id=id)

    # Default type *must* be set in order to delete any volume type.
    # If the default isn't set, the following call will raise
    # VolumeTypeDefaultMisconfiguredError exception which will error out the
    # delete operation.
    default_type = get_default_volume_type()
    # don't allow delete if the type requested is the conf default type
    if id == default_type.get('id'):
        raise exception.VolumeTypeDefaultDeletionError(volume_type_id=id)

    elevated = context if context.is_admin else context.elevated()
    return db.volume_type_destroy(elevated, id)


def get_all_types(context: context.RequestContext,
                  inactive: int = 0,
                  filters: Optional[dict] = None,
                  marker: Optional[dict[str, Any]] = None,
                  limit: Optional[int] = None,
                  sort_keys: Optional[list[str]] = None,
                  sort_dirs: Optional[list[str]] = None,
                  offset: Optional[int] = None,
                  list_result: bool = False) -> Union[dict[str, Any], list]:
    """Get all non-deleted volume_types.

    Pass true as argument if you want deleted volume types returned also.

    """
    vol_types = db.volume_type_get_all(context, inactive, filters=filters,
                                       marker=marker, limit=limit,
                                       sort_keys=sort_keys,
                                       sort_dirs=sort_dirs, offset=offset,
                                       list_result=list_result)
    return vol_types


def get_volume_type(
        ctxt: Optional[context.RequestContext],
        id: Optional[str],
        expected_fields: Optional[Iterable[str]] = None) -> dict[str, Any]:
    """Retrieves single volume type by id."""
    if id is None:
        msg = _("id cannot be None")
        raise exception.InvalidVolumeType(reason=msg)

    if ctxt is None:
        ctxt = context.get_admin_context()

    return db.volume_type_get(ctxt, id, expected_fields=expected_fields)


def get_by_name_or_id(context: context.RequestContext,
                      identity: str) -> dict[str, Any]:
    """Retrieves volume type by id or name"""
    if uuidutils.is_uuid_like(identity):
        # both name and id can be in uuid format
        try:
            return get_volume_type(context, identity)
        except exception.VolumeTypeNotFound:
            try:
                # A user can create a type with the name in a UUID format,
                # so here we check for the uuid-like name.
                return get_volume_type_by_name(context, identity)
            except exception.VolumeTypeNotFoundByName:
                raise exception.VolumeTypeNotFound(volume_type_id=identity)
    return get_volume_type_by_name(context, identity)


def get_volume_type_by_name(context: context.RequestContext,
                            name: Optional[str]) -> dict[str, Any]:
    """Retrieves single volume type by name."""
    if name is None:
        msg = _("name cannot be None")
        raise exception.InvalidVolumeType(reason=msg)

    return db.volume_type_get_by_name(context, name)


def get_default_volume_type(
        contxt: Optional[context.RequestContext] = None) -> dict[str, Any]:
    """Get the default volume type.

    :raises VolumeTypeDefaultMisconfiguredError: when the configured default
                                                 is not found
    """

    if contxt:
        project_default = db.project_default_volume_type_get(
            contxt, contxt.project_id)
        if project_default:
            return get_volume_type(contxt, project_default.volume_type_id)
    name = CONF.default_volume_type
    ctxt = context.get_admin_context()
    vol_type = {}
    try:
        vol_type = get_volume_type_by_name(ctxt, name)
    except (exception.VolumeTypeNotFoundByName, exception.InvalidVolumeType):
        # Couldn't find volume type with the name in default_volume_type
        # flag, record this issue and raise exception
        LOG.exception('Default volume type is not found. '
                      'Please check default_volume_type config:')
        raise exception.VolumeTypeDefaultMisconfiguredError(
            volume_type_name=name)

    return vol_type


def get_volume_type_extra_specs(
        volume_type_id: str,
        key: Union[str, bool] = False) -> Union[dict, bool]:
    volume_type = get_volume_type(context.get_admin_context(),
                                  volume_type_id)
    extra_specs = volume_type['extra_specs']
    if key:
        if extra_specs.get(key):
            return extra_specs.get(key)
        else:
            return False
    else:
        return extra_specs


def is_public_volume_type(context: context.RequestContext,
                          volume_type_id: str) -> bool:
    """Return is_public boolean value of volume type"""
    volume_type = db.volume_type_get(context, volume_type_id)
    return volume_type['is_public']


@utils.if_notifications_enabled
def notify_about_volume_type_access_usage(context: context.RequestContext,
                                          volume_type_id: str,
                                          project_id: str,
                                          event_suffix: str,
                                          host: Optional[str] = None) -> None:
    """Notify about successful usage type-access-(add/remove) command.

    :param context: security context
    :param volume_type_id: volume type uuid
    :param project_id: tenant uuid
    :param event_suffix: name of called operation access-(add/remove)
    :param host: hostname
    """
    notifier_info = {'volume_type_id': volume_type_id,
                     'project_id': project_id}

    if not host:
        host = CONF.host

    notifier = rpc.get_notifier("volume_type_project", host)
    notifier.info(context,
                  'volume_type_project.%s' % event_suffix,
                  notifier_info)


def add_volume_type_access(context: context.RequestContext,
                           volume_type_id: Optional[str],
                           project_id: str) -> None:
    """Add access to volume type for project_id."""
    if volume_type_id is None:
        msg = _("volume_type_id cannot be None")
        raise exception.InvalidVolumeType(reason=msg)
    elevated = context if context.is_admin else context.elevated()
    if is_public_volume_type(elevated, volume_type_id):
        msg = _("Type access modification is not applicable to public volume "
                "type.")
        raise exception.InvalidVolumeType(reason=msg)

    db.volume_type_access_add(elevated, volume_type_id, project_id)

    notify_about_volume_type_access_usage(context,
                                          volume_type_id,
                                          project_id,
                                          'access.add')


def remove_volume_type_access(context: context.RequestContext,
                              volume_type_id: Optional[str],
                              project_id: str) -> None:
    """Remove access to volume type for project_id."""
    if volume_type_id is None:
        msg = _("volume_type_id cannot be None")
        raise exception.InvalidVolumeType(reason=msg)
    elevated = context if context.is_admin else context.elevated()
    if is_public_volume_type(elevated, volume_type_id):
        msg = _("Type access modification is not applicable to public volume "
                "type.")
        raise exception.InvalidVolumeType(reason=msg)

    db.volume_type_access_remove(elevated, volume_type_id, project_id)

    notify_about_volume_type_access_usage(context,
                                          volume_type_id,
                                          project_id,
                                          'access.remove')


def is_encrypted(context: context.RequestContext,
                 volume_type_id: str) -> bool:
    return get_volume_type_encryption(context, volume_type_id) is not None


def get_volume_type_encryption(
        context: context.RequestContext,
        volume_type_id: Optional[str]) -> Optional[dict]:
    if volume_type_id is None:
        return None

    encryption = db.volume_type_encryption_get(context, volume_type_id)
    return encryption


def get_volume_type_qos_specs(volume_type_id: str) -> dict[str, Any]:
    """Get all qos specs for given volume type."""
    ctxt = context.get_admin_context()
    res = db.volume_type_qos_specs_get(ctxt,
                                       volume_type_id)
    return res


def volume_types_diff(context: context.RequestContext,
                      vol_type_id1: str,
                      vol_type_id2: str) -> tuple[dict[str, Any], bool]:
    """Returns a 'diff' of two volume types and whether they are equal.

    Returns a tuple of (diff, equal), where 'equal' is a boolean indicating
    whether there is any difference, and 'diff' is a dictionary with the
    following format:

    .. code-block:: default

        {
            'extra_specs': {'key1': (value_in_1st_vol_type,
                                     value_in_2nd_vol_type),
                            'key2': (value_in_1st_vol_type,
                                     value_in_2nd_vol_type),
                            {...}}
            'qos_specs': {'key1': (value_in_1st_vol_type,
                                   value_in_2nd_vol_type),
                          'key2': (value_in_1st_vol_type,
                                   value_in_2nd_vol_type),
                          {...}}
            'encryption': {'cipher': (value_in_1st_vol_type,
                                      value_in_2nd_vol_type),
                          {'key_size': (value_in_1st_vol_type,
                                        value_in_2nd_vol_type),
                           {...}}
        }
    """
    def _fix_qos_specs(qos_specs: Optional[dict]) -> None:
        if qos_specs:
            for key in QOS_IGNORED_FIELDS:
                qos_specs.pop(key, None)
            qos_specs.update(qos_specs.pop('specs', {}))

    def _fix_encryption_specs(encryption: Optional[dict]) -> Optional[dict]:
        if encryption:
            encryption = dict(encryption)
            for param in ENCRYPTION_IGNORED_FIELDS:
                encryption.pop(param, None)
        return encryption

    def _dict_diff(dict1: Optional[dict],
                   dict2: Optional[dict]) -> tuple[dict[str, Any], bool]:
        res = {}
        equal = True
        if dict1 is None:
            dict1 = {}
        if dict2 is None:
            dict2 = {}
        for k, v in dict1.items():
            res[k] = (v, dict2.get(k))
            if k not in dict2 or res[k][0] != res[k][1]:
                equal = False
        for k, v in dict2.items():
            res[k] = (dict1.get(k), v)
            if k not in dict1 or res[k][0] != res[k][1]:
                equal = False
        return (res, equal)

    all_equal = True
    diff = {}
    vol_type_data = []
    for vol_type_id in (vol_type_id1, vol_type_id2):
        if vol_type_id is None:
            specs = {'extra_specs': None,
                     'qos_specs': None,
                     'encryption': None}
        else:
            specs = {}
            vol_type = get_volume_type(context, vol_type_id)
            specs['extra_specs'] = vol_type.get('extra_specs')
            qos_specs = get_volume_type_qos_specs(vol_type_id)
            specs['qos_specs'] = qos_specs.get('qos_specs')
            _fix_qos_specs(specs['qos_specs'])
            specs['encryption'] = get_volume_type_encryption(context,
                                                             vol_type_id)
            specs['encryption'] = _fix_encryption_specs(specs['encryption'])
        vol_type_data.append(specs)

    diff['extra_specs'], equal = _dict_diff(vol_type_data[0]['extra_specs'],
                                            vol_type_data[1]['extra_specs'])
    if not equal:
        all_equal = False
    diff['qos_specs'], equal = _dict_diff(vol_type_data[0]['qos_specs'],
                                          vol_type_data[1]['qos_specs'])
    if not equal:
        all_equal = False
    diff['encryption'], equal = _dict_diff(vol_type_data[0]['encryption'],
                                           vol_type_data[1]['encryption'])
    if not equal:
        all_equal = False

    return (diff, all_equal)


def volume_types_encryption_changed(
        context: context.RequestContext,
        vol_type_id1: str, vol_type_id2: str) -> bool:
    """Return whether encryptions of two volume types are same."""
    def _get_encryption(enc: dict) -> dict:
        enc = dict(enc)
        for param in ENCRYPTION_IGNORED_FIELDS:
            enc.pop(param, None)
        return enc

    enc1 = get_volume_type_encryption(context, vol_type_id1)
    enc2 = get_volume_type_encryption(context, vol_type_id2)

    enc1_filtered = _get_encryption(enc1) if enc1 else None
    enc2_filtered = _get_encryption(enc2) if enc2 else None
    return enc1_filtered != enc2_filtered


def provision_filter_on_size(context: context.RequestContext,
                             volume_type: Optional[dict[str, Any]],
                             size: Union[str, int]) -> None:
    """This function filters volume provisioning requests on size limits.

    If a volume type has provisioning size min/max set, this filter
    will ensure that the volume size requested is within the size
    limits specified in the volume type.
    """

    if not volume_type:
        volume_type = get_default_volume_type()

    if volume_type:
        size_int = int(size)
        extra_specs = volume_type.get('extra_specs', {})
        min_size = extra_specs.get(MIN_SIZE_KEY)
        if min_size and size_int < int(min_size):
            msg = _("Specified volume size of '%(req_size)d' is less "
                    "than the minimum required size of '%(min_size)s' "
                    "for volume type '%(vol_type)s'.") % {
                'req_size': size_int, 'min_size': min_size,
                'vol_type': volume_type['name']
            }
            raise exception.InvalidInput(reason=msg)

        max_size = extra_specs.get(MAX_SIZE_KEY)
        if max_size and size_int > int(max_size):
            msg = _("Specified volume size of '%(req_size)d' is "
                    "greater than the maximum allowable size of "
                    "'%(max_size)s' for volume type '%(vol_type)s'."
                    ) % {
                'req_size': size_int, 'max_size': max_size,
                'vol_type': volume_type['name']}
            raise exception.InvalidInput(reason=msg)
